from ride import lifecycle  # noqa: F401


class TestClassificationLifecycle:
    def test_validate_attributes(self):
        """Test validation of attributes"""
        assert False

    def test_metrics_step(self):
        assert False

    def test_metrics_epoch(self):
        assert False

    def test_training_step(self):
        assert False

    def test_training_epoch_end(self):
        assert False

    def test_validation_step(self):
        assert False

    def test_validation_epoch_end(self):
        assert False

    def test_test_step(self):
        assert False

    def test_test_epoch_end(self):
        assert False
