from ride import hparamsearch  # noqa: F401


def test_add_argparse_args():
    assert False


def test_run():
    assert False


def test_call():
    assert False


def test_dump():
    assert False


def test_load():
    assert False
